from hedra.plugins.types.extension import ExtensionPlugin
from typing import Dict, Union


def get_enabled_extensions() -> Dict[str, Union[ExtensionPlugin, None]]:

    enabled_extensions = {}
    
    return enabled_extensions