from .generated_graphql_execute_stage import ExecuteGraphQLStage
from .generated_graphql_http2_stage import ExecuteGraphQLHttp2Stage
from .generated_http_execute_stage import ExecuteHTTPStage
from .generated_http2_execute_stage import ExecuteHTTP2Stage
from .generated_http3_execute_stage import ExecuteHTTP3Stage
from .generated_playwright_execute_stage import ExecutePlaywrightStage
from .generated_task_execute_stage import ExecuteTaskStage
from .generated_udp_execute_stage import ExecuteUDPStage
from .generated_websocket_execute_stage import ExecuteWebsocketStage