from enum import Enum


class MutationType(Enum):
    DEFORM_HEADER='DEFORM_HEADER'
    INJECT_HEADER='INJECT_HEADER'
    INJECT_JUNK_DATA='INJECT_JUNK_DATA'
    INJECT_PING='INJECT_PING'
    SMUGGLE_REQUEST='SMUGGLE_REQUEST'